# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json

import gear

from gearhorn import worker


class GearhornClient(gear.Client):
    '''An example of a client object to use to subscribe to Gearhorn
    broadcasts.'''

    def submitJob(self, job, background=False, timeout=30):
        real_name = job.name
        real_arguments = {'topic': real_name, 'payload': job.arguments}
        if background:
            real_arguments['background'] = True
        job.name = worker.GearhornWorker.fanout_name
        job.arguments = bytes(json.dumps(real_arguments).encode('utf-8'))
        return super(
            GearhornClient, self).submitJob(job, background=background,
                                            timeout=timeout)



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

import gear

from gearhorn.tests import base
from gearhorn.tests import server
from gearhorn import subscriber
from gearhorn import worker


class TestGearhornSubscriber(base.TestCase):
    def setUp(self):
        super(TestGearhornSubscriber, self).setUp()
        self.server = server.TestServer()
        self.addCleanup(self.server.shutdown)
        self.client = gear.Client('in_client')
        self.addCleanup(self.client.shutdown)
        self.client.addServer('localhost', self.server.port)
        self.client.waitForServer()

    def test_subscriber(self):
        w = gear.Worker('test_subscriber_worker')
        w.registerFunction(worker.GearhornWorker.subscribe_name)
        self.addCleanup(w.shutdown)
        w.addServer('localhost', self.server.port)
        s = subscriber.GearhornSubscriber('test_subscriber')
        s.addServer('localhost', self.server.port)
        s.waitForServer()
        s.subscribe('a_topic')

        def _assertJob():
            p = json.loads(j.arguments)
            self.assertIsInstance(p, dict)
            self.assertIn('topic', p)
            self.assertIn('client_id', p)
            self.assertEqual('a_topic', p['topic'])
            self.assertEqual('test_subscriber', p['client_id'])

        j = w.getJob()
        _assertJob()
        w.registerFunction(worker.GearhornWorker.unsubscribe_name)
        s.unsubscribe('a_topic')
        j = w.getJob()
        _assertJob()



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import socket

from gearhorn import worker
from gearhorn.store import sqla


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('host', default=['localhost'],
                        help='Gearman server(s)', nargs='*')
    parser.add_argument('--sqlalchemy-dsn', help='SQLAlchemy DSN for storing'
                        ' matchmaking data')
    opts = parser.parse_args()
    w = worker.GearhornWorker(client_id='gearhorn_%s' % socket.gethostname(),
                              dsn=opts.sqlalchemy_dsn)
    for host in opts.host:
        if '/' in host:
            (host, port) = host.split('/')
            w.addServer(host, port=port)
        else:
            w.addServer(host)
    try:
        while True:
            w.work()
    except Exception as e:
        print(str(e))
        return -1
    return 0

def init_schema():
    parser = argparse.ArgumentParser()
    parser.add_argument('sqlalchemy_dsn', help='SQLAlchemy DSN for schema '
                        'to initialize.')
    opts = parser.parse_args()
    s = sqla.Store(opts.sqlalchemy_dsn)
    s.initialize_schema()



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random

import gear

class TestServer(gear.Server):
    ''' Selects a random port to listen on '''
    def __init__(self):
        port = random.randint(30000, 31000)
        while True:
            try:
                super(TestServer, self).__init__(port=port)
                break
            except Exception as e:
                if 'could not open socket' in str(e):
                    print('failed to listen on %d' % port)
                    pass
                raise e



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sqlalchemy as sa
from sqlalchemy import exc as sa_exc
from sqlalchemy import orm
from sqlalchemy.orm import exc

from gearhorn.store import sqla_models as models
from gearhorn import util

_session = None


class Store(object):
    def __init__(self, details):
        self._engine = sa.create_engine(details)
        self._sessionmaker = orm.sessionmaker(bind=self._engine)

    @property
    def session(self):
        return orm.scoped_session(self._sessionmaker)

    def initialize_schema(self):
        models.Base.metadata.create_all(self._engine)

    def subscribe(self, client_id, topic):
        # Storing as binary for more efficient usage in MySQL particularly
        s = models.Subscriber(topic=util.to_utf8(topic),
                              client_id=util.to_utf8(client_id))
        sess = self.session()
        try:
            sess.add(s)
            # XXX: allow commit in batches some day
            sess.commit()
        except sa_exc.IntegrityError:
            sess.rollback()
            # We don't actually care, because it means we already have this
            # subscription
            pass

    def unsubscribe(self, client_id, topic):
        s = models.Subscriber(topic=util.to_utf8(topic),
                              client_id=util.to_utf8(client_id))
        sess = self.session()
        sess.delete(s)
        sess.commit()

    def get_subscribers(self, topic):
        sess = self.session()
        try:
            for sub in sess.query(
                models.Subscriber.client_id).filter_by(topic=topic):
                yield sub[0]
        except exc.NoResultFound:
            return



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json

import gear

from gearhorn import worker


def get_broadcast_function(client_id, topic):
    return '%s_%s' % (topic, client_id)


class GearhornSubscriber(object):
    '''Subscribe to broadcasts from Gearhorn'''

    def __init__(self, client_id):
        self.client_id = client_id
        self.worker = gear.Worker(client_id=client_id)
        client_client_id = '%s_subscriber' % (client_id)
        self.client = gear.Client(client_id=client_client_id)

    def _make_sub_job(self, name, topic):
        message = {'client_id': self.client_id,
                   'topic': topic}
        return gear.Job(name=name,
                        arguments=bytes(json.dumps(message).encode('utf-8')))

    def addServer(self, host, port=4730, ssl_key=None, ssl_cert=None,
                  ssl_ca=None):
        self.worker.addServer(host, port=port, ssl_key=ssl_key,
                              ssl_cert=ssl_cert, ssl_ca=ssl_ca)
        self.client.addServer(host, port=port, ssl_key=ssl_key,
                              ssl_cert=ssl_cert, ssl_ca=ssl_ca)

    def waitForServer(self):
        self.worker.waitForServer()
        self.client.waitForServer()

    def subscribe(self, topic):
        self.worker.registerFunction(
            get_broadcast_function(self.client_id, topic))
        self.client.submitJob(
            self._make_sub_job(worker.GearhornWorker.subscribe_name, topic))

    def unsubscribe(self, topic):
        self.client.submitJob(
            self._make_sub_job(worker.GearhornWorker.unsubscribe_name, topic))
        self.worker.unRegisterFunction(
            get_broadcast_function(self.client_id, topic))



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import time

import gear

from gearhorn import util
from gearhorn.store import sqla


class GearhornWorker(gear.Worker):
    subscribe_name = 'subscribe_fanout'
    unsubscribe_name = 'unsubscribe_fanout'
    fanout_name = 'fanout'
    foreground_timeout = 30

    def __init__(self, client_id=None, worker_id=None, dsn=None):
        super(GearhornWorker, self).__init__(client_id=client_id,
                                             worker_id=worker_id)
        client_client_id = (client_id or worker_id) + '_broadcaster'
        self._servers = []
        self.client = gear.Client(client_id=client_client_id)
        self._store = sqla.Store(dsn)

    def addServer(self, host, port=4730, ssl_key=None, ssl_cert=None,
                  ssl_ca=None):
        super(GearhornWorker, self).addServer(host, port, ssl_key, ssl_cert,
                                              ssl_ca)
        self.client.addServer(host, port, ssl_key, ssl_cert, ssl_ca)

    def registerSubscriberFunctions(self):
        return self.registerFunction(self.subscribe_name)
        return self.registerFunction(self.unsubscribe_name)

    def registerFanoutFunction(self):
        return self.registerFunction(self.fanout_name)

    def work(self):
        job = self.getJob()
        if job.name == self.fanout_name:
            return self.fanout(job)
        elif job.name == self.subscribe_name:
            return self.subscribe(job)
        elif job.name == self.unsubscribe_name:
            return self.unsubscribe(job)
        raise RuntimeError('Unknown job %s' % job.name)

    def subscribe(self, job):
        self._subunsub(job, self._store.subscribe)

    def unsubscribe(self, job):
        self._subunsub(job, self._store.unsubscribe)

    def _subunsub(self, job, action):
        try:
            message = json.loads(job.arguments)
            if not isinstance(message, dict):
                raise ValueError('must be a mapping')
            if 'topic' not in message or 'client_id' not in message:
                raise ValueError('must have topic and client_id keys')
        except ValueError as e:
            self.log.debug(e, exc_info=True)
            job.sendWorkException(bytes(str(e).encode('utf-8')))
            return
        self.log.debug('Attempting %s %s' % (action, message))
        try:
            action(client_id=message['client_id'],
                   topic=message['topic'])
        except Exception as e:
            self.log.debug(e, exc_info=True)
            job.sendWorkException(bytes(str(e).encode('utf-8')))
            return
        self.log.info('Completed %s %s' % (action, message))
        job.sendWorkComplete()

    def fanout(self, job):
        try:
            message = json.loads(job.arguments)
            if not isinstance(message, dict):
                raise ValueError('must be a JSON mapping.')
            if 'topic' not in message or 'payload' not in message:
                raise ValueError('must have topic and payload keys')
        except ValueError as e:
            job.sendWorkException(bytes(str(e).encode('utf-8')))
            return
        wait_jobs = []
        errors = []
        for sub in self._store.get_subscribers(message['topic']):
            name = '%s_%s' % (message['topic'], sub)
            unique = util.to_utf8(message.get('unique'))
            cj = gear.Job(util.to_utf8(name),
                          arguments=util.to_utf8(message['payload']),
                          unique=unique)
            try:
                self.client.submitJob(cj, background=message.get('background',
                                                                 False))
                if not message.get('background'):
                    wait_jobs.append((sub, cj))
            except gear.GearmanError as e:
                errors.append((sub, str(e)))
        done = 0
        # Timeout just in case
        before = time.time()
        while done < len(wait_jobs):
            for sub, wj in wait_jobs:
                if wj.complete and wj.failure:
                    if wj.exception:
                        errors.append((sub, wj.exception))
                    else:
                        errors.append((sub, 'Worker failure'))
                elif wj.complete:
                    done += 1
            time.sleep(0.1)
            if time.time() - before > self.foreground_timeout:
                # timed out
                for sub, wj in wait_jobs:
                    if not wj.complete:
                        errors.append((sub, 'Worker timeout'))
                break
        if errors:
            job.sendWorkException(json.dumps(errors))
        else:
            job.sendWorkComplete(bytes('%d' % done).encode('utf-8'))
        return True



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import threading
import time

import fixtures
import gear

from gearhorn.tests import base
from gearhorn.tests import server
from gearhorn import worker


class TestGearhornWorker(base.TestCase):
    def setUp(self):
        super(TestGearhornWorker, self).setUp()
        self.server = server.TestServer()
        self.addCleanup(self.server.shutdown)
        self.client = gear.Client('in_client')
        self.addCleanup(self.client.shutdown)
        self.client.addServer('localhost', self.server.port)
        self.client.waitForServer()

    def test_worker(self):
        tdir = self.useFixture(fixtures.TempDir())
        dsn = 'sqlite:///%s/test.db' % tdir.path
        w = worker.GearhornWorker(client_id='test_worker', dsn=dsn)
        w._store.initialize_schema()
        self.addCleanup(w.shutdown)
        w.addServer('localhost', self.server.port)
        w.registerSubscriberFunctions()
        w.registerFanoutFunction()
        subw = gear.Worker('test_worker_subw')
        subw.addServer('localhost', self.server.port)
        subw.waitForServer()
        subw.registerFunction('broadcasts_test_receiver')
        subscribe_message = {'client_id': 'test_receiver',
                             'topic': 'broadcasts'}
        subscribe_job = gear.Job(w.subscribe_name,
                                 arguments=json.dumps(
                                     subscribe_message).encode('utf-8'))
        self.client.submitJob(subscribe_job)
        # w should have this message only
        w.work()
        while not subscribe_job.complete:
            time.sleep(0.1)
        fanout_message = {'topic': 'broadcasts',
                          'payload': 'in payload'}
        job = gear.Job(w.fanout_name,
                       json.dumps(fanout_message).encode('utf-8'))
        self.client.submitJob(job)
        # Thread in the background to wait for subw to complete job
        t = threading.Thread(target=w.work)
        t.start()
        broadcasted = subw.getJob()
        self.assertEqual('in payload', broadcasted.arguments)
        broadcasted.sendWorkComplete()
        # wait for complete to wind through tubes
        while not job.complete:
            time.sleep(0.1)
        self.assertFalse(job.failure)
        self.assertIsNone(job.exception)
        self.assertEqual([b'1'], job.data)
        t.join()
